"""Container for commonly shared constants."""

SAMPLERATE_SPS_DEFAULT: int = 100_000

UID_NAME: str = "SHEPHERD_NODE_ID"
UID_SIZE: int = 2

TESTBED_SERVER_URI: str = "http://127.0.0.1:8000/shepherd"
